#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <list>
using namespace std;

extern "C"
__declspec(dllimport)
int __stdcall MultiByteToWideChar(int, int, const char*, int, wchar_t*, int);


bool g_pal;
int g_res_x, g_res_y;


// I basically had to reimplement this because the standard sscanf is defined
// not to accept zero-length strings for the %[...] notation. This is
// incredibly brain-damaged because there's no way (that I can see) to
// recover the behavior I want from this, whereas it would be easy to go the
// other way (just check afterwards for an empty string).

bool my_sscanf(const char* s, const char* fmt, ...) {
	va_list val;
	va_start(val, fmt);
	for (const char* p = fmt; *p; ++p) {
		if (*p != '%') {
			if (*p != *s) return false;
			++s;
		} else {
			++p;
			int count=0;
			while (*p >= '0' && *p <= '9')
				count = count*10 + (*p++ - '0');
			switch (*p) {
				case 'd': case 'i': case 'X':
				{
					int* arg = va_arg(val, int*);
					const char* old_s = s;
					*arg = strtol(s, (char**)&s, (*p == 'd' ? 10 : *p == 'X' ? 16 : 0));
					if (s == old_s) return false;
					break;
				}
				case '[':
				{
					if (p[1] != '^' || p[3] != ']') {
						return false;
					}
					char terminator = p[2];
					p += 3;
					char* arg = va_arg(val, char*);
					if (count == 0) count = -1;
					while (*s != terminator && *s != 0) {
						*arg++ = *s++;
						if (count-- == 0) return false;
					}
					*arg = 0;
					break;
				}
				default:
					return false;
			}
		}
	}
	va_end(val);
	return true;
}


unsigned CvtSSAColor(unsigned color) {
	return 0xF000000 + ((color & 0x0000FF) << 16) + (color & 0x00FF00) + ((color & 0xFF0000) >> 16);
}

unsigned CvtSSATime(unsigned h, unsigned m, unsigned s, unsigned hund) {
	unsigned t = ((h*60+m)*60+s)*100+hund;
	if (g_pal)
		return t / 2;
	else
		return (t * 600 + 500) / 1001;
}


struct SSAStyle {
	SSAStyle* next;
	char name[50];
	char typeface[50];
	int typesize;
	int textcolor, secondarycolor, tertiarycolor, halocolor;
	int bold, italic;
	int shadow, shadow_width;
	int align;
	int lmargin, rmargin, vmargin;
	int alpha;
};


SSAStyle default_style = {
	0,
	"",
	"Arial",
	27,
	0x00FFFF, 0xFFFFFF, 0x00FF00, 0,
	-1, 0,
	0, 0,
	2,
	30, 30, 30,
	0
};


void PrintTextBox(int lmargin, int rmargin, int vmargin, int align, FILE* out) {
	if (g_res_x == 0) {
		g_res_x = 640;
		g_res_y = 480;
	}
	int hboxalign = (align&3)-1;
	int vboxalign = (align<4) ? 0 : (align<8) ? 2 : 1;
	int left   = lmargin;
	int right  = g_res_x - rmargin;
	int top    = vboxalign==2 ? vmargin : 0;
	int bottom = vboxalign==0 ? g_res_y - vmargin : g_res_y;
	left = (left * 720 + g_res_x/2) / g_res_x;
	right = (right * 720 + g_res_x/2) / g_res_x;
	top = (top * (g_pal ? 576 : 480) + g_res_y/2) / g_res_y;
	bottom = (bottom * (g_pal ? 576 : 480) + g_res_y/2) / g_res_y;
	fprintf(out, "<textbox %03d,%03d,%03d,%03d,%c%c>",
		left, right, top, bottom,
		vboxalign*3 + hboxalign + '1',
		"LCR"[hboxalign]);
}


void PrintSSAStyleAsMacro(SSAStyle* s, FILE* out) {
	int size = s->typesize;
	if (g_res_y) size = size * (g_pal ? 576 : 480) / g_res_y;
	fprintf(out, "<def %s><font %s><size %d><color %X,%X><%sb><%si>",
		s->name, s->typeface, size, CvtSSAColor(s->textcolor), CvtSSAColor(s->halocolor),
		s->bold ? "" : "/", s->italic ? "" : "/");
	PrintTextBox(s->lmargin, s->rmargin, s->vmargin, s->align, out);
	fprintf(out, "<enddef>\n");
}


const wchar_t UNICODE_REPLACEMENT = 0xFFFD;

void PrintText(const char* text, const char* text_end, FILE* out) {
	unsigned len = MultiByteToWideChar(0/*FIXME*/, 0, text, text_end-text, 0, 0);
	wchar_t* wtext = new wchar_t[len+1];
	len = MultiByteToWideChar(0/*FIXME*/, 0, text, text_end-text, wtext, len+1);
	wtext[len] = 0;
	for (int i=0; i<len; ++i) {
		wchar_t c = wtext[i];
		if (c >= 0xD800 && c < 0xE000) {
			wchar_t d = wtext[i+1];
			if (c < 0xDC00 && d >= 0xDC00 && d < 0xE000) {
				c = (c - 0xD800) * 1024 + (d - 0xDC00) + 0x10000;
				++i;
			} else {
				c = UNICODE_REPLACEMENT;
			}
		}
		if (c < 0x80) {
			putc(c, out);
		} else {
			int n;
			if (c < 0x800)
				n = 1;
			else if (c < 0x10000)
				n = 2;
			else
				n = 3;
			putc((0xFF << (7-n)) + (c >> (n*6)), out);
			for (int x=0; x<n; ++x) {
				putc(0x80 + ((c >> ((n-1-x)*6)) & 63), out);
			}
		}
	}
	delete[] wtext;
}


void Convert(FILE* in, FILE* out) {
	char line[1024];
	int version = 2;
	SSAStyle* ssa_styles = 0;
	fputs("# Converted by ssa2dvdsub\n\n# You will need to set these three options manually.\n<vts 01>\n<basefield 0>\n<angle 1>\n\n", out);
	while (fgets(line, 1024, in)) {
		if (memcmp(line, "Dialogue: ", 10) == 0 || memcmp(line, "Comment: ", 9) == 0) {
			int marked,h1,m1,s1,f1,h2,m2,s2,f2,lmargin,rmargin,vmargin;
			char style[50],speaker[50];
			char dialogue[1024];
			// FIXME: effect field for v3
			bool success = my_sscanf(line+(line[0]=='C'?9:10), "Marked=%d,%d:%d:%d.%d,%d:%d:%d.%d,%49[^,],%49[^,],%d,%d,%d,%[^\n]",
				&marked, &h1,&m1,&s1,&f1,&h2,&m2,&s2,&f2, style, speaker, &lmargin, &rmargin, &vmargin, dialogue);
			if (!success) {
				fprintf(stderr, "Dialogue line format unrecognized (ignoring):\n%s", line);
			} else {
				if (version < 3) {
					f1 *= 10;
					f2 *= 10;
				}
				// find style
				SSAStyle* s;
				for (s = ssa_styles; s; s = s->next) {
					if (strcmp(s->name, style) == 0)
						break;
				}
				if (!s) {
					fprintf(stderr, "Style \"%s\" not found; using default\n", style);
					s = new SSAStyle(default_style);
					s->next = ssa_styles;
					strcpy(s->name, style);
					ssa_styles = s;
					PrintSSAStyleAsMacro(s, out);
				}
				if (line[0] == 'C')
					putc('#', out);
				fprintf(out, "<at %06d,%06d><%s>", CvtSSATime(h1,m1,s1,f1), CvtSSATime(h2,m2,s2,f2), style);
				int align = 0;
				char* p = dialogue;
				if (version >= 3) {
					// hack: this version has an extra field; skip over it by hand
					while (*p != ',' && *p != 0) ++p;
					if (*p != 0) *p++ = 0;
					if (dialogue[0] != 0) {
						fprintf(stderr, "Effect not supported (%s); ignoring\n", dialogue);
					}
				}
				bool first_text = true;
				while (*p) {
					if (p[0] == '\\' && (p[1] == 'n' || p[1] == 'N')) {
						putc('|', out);
						p += 2;
					} else if (p[0] == '{') {
						++p;
						while (*p != '}') {
							if (*p != '\\') {
								fprintf(stderr, "Broken escape sequence:\n%s", line);
								break;
							}
							++p;
							if (p[0] == 'f' && p[1] == 's') {
								p += 2;
								int size = strtol(p, &p, 10);
								if (g_res_y) size = size * (g_pal ? 576 : 480) / g_res_y;
								fprintf(out, "<size %d>", size);
							} else if (p[0] == 'f' && p[1] == 'n') {
								p += 2;
								char* name = p;
								while (*p != '\\' && *p != '}' && *p != 0) {
									++p;
								}
								char save = *p;
								*p = 0;
								fprintf(out, "<font %s>", name);
								*p = save;
							} else if (p[0] == 'i') {
								fputs(p[1] == '0' ? "</i>" : "<i>", out);
								p += 2;
							} else if (p[0] == 'b') {
								fputs(p[1] == '0' ? "</b>" : "<b>", out);
								p += 2;
							} else if (p[0] == 'c' && p[1] == '&' && p[2] == 'H') {
								p += 3;
								unsigned color = 0;
								my_sscanf(p, "%X", &color);
								printf("<color %X,%X>", CvtSSAColor(color), CvtSSAColor(s->halocolor));
								while (*p != '&' && *p != 0)
									++p;
								if (*p == '&')
									++p;
							} else if (p[0] == 'a') {
								p += 1;
								align = strtol(p, &p, 10);
							} else {
								fprintf(stderr, "unrecognized escape code:\n%s", line);
								while (*p != '\\' && *p != '}' && *p != 0)
									++p;
							}
						}
						if (*p == '}')
							++p;
					} else {
						if (first_text && (lmargin || rmargin || vmargin || align)) {
							PrintTextBox(
								lmargin ? lmargin : s->lmargin,
								rmargin ? rmargin : s->rmargin,
								vmargin ? vmargin : s->vmargin,
								align ? align : s->align, out);
						}
						first_text = false;
						char* text = p;
						while (*p != '{' && *p != '\\' && *p != 0) {
							++p;
						}
						PrintText(text, p, out);
					}
				}
				putc('\n', out);
			}
		} else if (memcmp(line, "Style: ", 7) == 0) {
			SSAStyle* s = new SSAStyle;
			// try to determine format by counting the number of commas
			int num_fields = 1;
			for (const char* p = line+7; *p; ++p)
				num_fields += (*p == ',');
			bool success = false;

			int dummy_encoding, dummy_border;

			switch (num_fields) {
			case 16:
				success = my_sscanf(line+7, "%49[^,],%49[^,],%d,%i,%i,%i,%i,%d,%d,%d,%d,%d,%d,%d,%d,%d",
					s->name, s->typeface, &s->typesize,
					&s->textcolor, &s->secondarycolor, &s->tertiarycolor, &s->halocolor,
					&s->bold, &s->italic, &dummy_border, &dummy_border, /*&s->shadow, &s->shadow_width,*/
					&s->align, &s->lmargin, &s->rmargin, &s->vmargin, &s->alpha);
				break;
			case 18:
				success = my_sscanf(line+7, "%49[^,],%49[^,],%d,%i,%i,%i,%i,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
					s->name, s->typeface, &s->typesize,
					&s->textcolor, &s->secondarycolor, &s->tertiarycolor, &s->halocolor,
					&s->bold, &s->italic, &dummy_border, &dummy_border, &dummy_border,
					&s->align, &s->lmargin, &s->rmargin, &s->vmargin, &s->alpha, &dummy_encoding);
			}
			if (!success) {
				fprintf(stderr, "Style line format unrecognized (ignoring) (%d):\n%s", num_fields, line);
			} else {
				s->next = ssa_styles;
				ssa_styles = s;
				PrintSSAStyleAsMacro(s, out);
			}
		} else if (line[0] == '[' && version >= 4) {
			// ignore
		} else if (line[0] == ';' || (line[0] == '!' && line[1] == ':')) {
			// ignore (should these be included in the output?)
		} else if (memcmp(line, "ScriptType: v", 13) == 0) {
			version = line[13] - '0';
		} else if (memcmp(line, "PlayResY: ", 10) == 0) {
			int y;
			if (!my_sscanf(line+10, "%d", &y)) {
				fprintf(stderr, "Line format unrecognized (ignoring):\n%s", line);
			} else {
				g_res_y = y;
				if (g_res_x == 0) {
					g_res_x = g_res_y * 4 / 3;
				}
			}
		} else if (memcmp(line, "PlayResX: ", 10) == 0) {
			int x;
			if (!my_sscanf(line+10, "%d", &x)) {
				fprintf(stderr, "Line format unrecognized (ignoring):\n%s", line);
			} else {
				g_res_x = x;
				if (g_res_y == 0) {
					g_res_y = g_res_x * 3 / 4;
				}
			}
		} else {
			char* newline = strchr(line, '\n');
			if (newline) *newline = 0;
			fprintf(out, "## %s\n", line);
		}
	}
}


int main(int argc, char** argv) {
	if (argc > 1 && strcmp(argv[1], "-p") == 0)
		g_pal = true;
	Convert(stdin, stdout);
	return 0;
}
